# -*- coding: utf-8 -*-

import os
import sys
from datetime import datetime
from typing import Optional, List, Tuple
from enum import Enum

import psutil

from PyQt5 import uic
from PyQt5.QtCore import Qt, QTimer, QProcess
from PyQt5.QtGui import QGuiApplication, QIcon, QColor, QPalette, QKeyEvent
from PyQt5.QtWidgets import QApplication, QMessageBox, QMainWindow, QFileDialog, QTabWidget, QLabel, \
    QPushButton, QTextBrowser, QComboBox, QCheckBox, QLineEdit

from astra_csp.cprocsp_process_dialog import CprocspProcessDialog
from astra_csp.interactive_process_dialog import InteractiveProcessDialog
from astra_csp.certificate_process import Certificate
from astra_csp.certificate_process import CertificateProcess
from astra_csp.create_file_signature_process import CreateFileSignatureProcess
from astra_csp.pdf_conversion_process import PdfConversionProcess
from astra_csp.verify_file_process import VerifyFileProcess
from astra_csp.application_cache import ApplicationCache
from astra_csp.settings import GlobalSettings
from astra_csp.settings import UserSettings
from astra_csp.utils import Utils

translate = QApplication.translate

mainwindow_ui = Utils.file_path('ui', 'mainwindow.ui')
UiFormClass, UiFormBase = uic.loadUiType(mainwindow_ui)  # type: QMainWindow, QMainWindow


class ResultStatus(Enum):
    Normal = 0
    Warning = 1
    Error = 2


class MainWindow(UiFormClass, UiFormBase):
    def __init__(self) -> None:
        super(MainWindow, self).__init__()
        self.tabWidget = None  # type: Optional[QTabWidget]
        self.labelDocumentPath = None  # type: Optional[QLabel]
        self.pushButtonDocumentView = None  # type: Optional[QPushButton]
        self.labelSignaturePath = None  # type: Optional[QLabel]
        self.pushButtonVerifySignature = None  # type: Optional[QPushButton]
        self.pushButtonSelectSignature = None  # type: Optional[QPushButton]
        self.textBrowserResult = None  # type: Optional[QTextBrowser]
        self.labelDocumentPathSign = None  # type: Optional[QLabel]
        self.pushButtonDocumentViewSign = None  # type: Optional[QPushButton]
        self.comboBoxCertificates = None  # type: Optional[QComboBox]
        self.checkBoxIncludeTimestamp = None  # type: Optional[QCheckBox]
        self.lineEditSignatureFileName = None  # type: Optional[QLineEdit]
        self.pushButtonSign = None  # type: Optional[QPushButton]
        self.pushButtonGenerateName = None  # type: Optional[QPushButton]
        self.textBrowserResultSign = None  # type: Optional[QTextBrowser]
        self.setupUi(self)

        self.pushButtonDocumentView.clicked.connect(self.view)
        self.pushButtonDocumentViewSign.clicked.connect(self.view)
        self.pushButtonSelectSignature.clicked.connect(self.select)
        self.pushButtonVerifySignature.clicked.connect(self.verify)
        self.pushButtonSign.clicked.connect(self.sign)
        self.pushButtonGenerateName.clicked.connect(self.update_signature_name)
        self.lineEditSignatureFileName.textChanged.connect(self.clear_sign_result)
        self.comboBoxCertificates.currentIndexChanged.connect(self.clear_sign_result)
        self.checkBoxIncludeTimestamp.stateChanged.connect(self.clear_sign_result)

        self.success_palette = QGuiApplication.palette()
        self.success_palette.setColor(QPalette.Base, QColor("#DDFFDE"))
        self.success_palette.setColor(QPalette.Text, QColor("#000000"))
        self.warning_palette = QGuiApplication.palette()
        self.warning_palette.setColor(QPalette.Base, QColor("yellow"))
        self.warning_palette.setColor(QPalette.Text, QColor("#000000"))
        self.error_palette = QGuiApplication.palette()
        self.error_palette.setColor(QPalette.Base, QColor("#FE4543"))
        self.error_palette.setColor(QPalette.Text, QColor("#FFFFFF"))

        fallbackicon = QIcon.fromTheme('document-edit-sign')
        icon = QIcon.fromTheme('document-sign', fallbackicon)
        self.setWindowIcon(icon)
        self.tabWidget.setCurrentIndex(0)

        self.files = []  # type: List[str]
        self.document = None  # type: Optional[str]
        self.signature = None  # type: Optional[str]
        self.user_settings = None  # type: Optional[UserSettings]
        self.global_settings = None  # type: Optional[GlobalSettings]
        self.sigmimetypes = ['application/pkcs7-signature', 'application/pgp-signature']
        self.certificates_process = CertificateProcess()
        self.create_file_signature_process = CreateFileSignatureProcess()
        self.pdf_conversion_process = PdfConversionProcess()
        self.verify_file_process = VerifyFileProcess()
        self.cache = ApplicationCache()

        self.comboBoxCertificates.setCurrentText(translate('text', 'Loading...'))

        self.waiting_dialog = None  # type: Optional[InteractiveProcessDialog]
        self.waiting_timer = QTimer(self)
        self.waiting_timer.setSingleShot(True)
        self.waiting_timer.setInterval(2000)
        self.waiting_timer.timeout.connect(self.show_waiting_dialog)

    # noinspection PyPep8Naming
    def closeEvent(self, event):
        self.clear_cache()
        event.accept()

    # noinspection PyPep8Naming
    def keyPressEvent(self, keyevent: QKeyEvent) -> None:
        if keyevent.key() == Qt.Key_F1:
            self.show_help()

    def initialize(self, files: List[str]) -> None:
        try:
            documents, signatures = self.parse_documents_and_signatures(files)
            if documents:
                self.document = documents[0]
            if signatures:
                self.signature = signatures[0]
            self.validate_file()
            self.validate_signature()
            self.user_settings = UserSettings()
            self.global_settings = GlobalSettings()
            self.validate_system()
            self.certificates_process.kill_time = self.global_settings.max_csp_time_waiting
            self.create_file_signature_process.kill_time = self.global_settings.max_csp_time_waiting
            self.pdf_conversion_process.kill_time = self.global_settings.max_csp_time_waiting
            self.verify_file_process.kill_time = self.global_settings.max_csp_time_waiting
            self.certificates_process.finished.connect(self.on_get_certificates_process_finished)
            self.create_file_signature_process.finished.connect(self.on_create_file_signature_process_finished)
            self.pdf_conversion_process.finished.connect(self.on_pdf_conversion_process_finished)
            self.verify_file_process.finished.connect(self.on_verify_file_process_finished)

            self.start_waiting(CprocspProcessDialog(process=self.certificates_process, parent=self))
            self.certificates_process.start()
        except Exception as exception:
            self.show_critical_error(translate('text', 'Failed to initialize system'), details=str(exception))

    def on_get_certificates_process_finished(self) -> None:
        self.stop_waiting()
        if self.certificates_process.exitStatus() != QProcess.NormalExit or \
           self.certificates_process.exitCode() != 0:
            self.show_critical_error(self.certificates_process.errorString(),
                                     translate('text', 'Failed to load certificates'))
            return

        self.setup_certificates(self.certificates_process.get_certificates())
        self.enable_ui()
        self.maybe_verify()

    def on_create_file_signature_process_finished(self) -> None:
        self.stop_waiting()
        if self.create_file_signature_process.exitStatus() != QProcess.NormalExit:
            self.show_error(self.create_file_signature_process.errorString(),
                            translate('text', 'Failed to create file signature'))
            return

        if self.create_file_signature_process.timestamp_error_occurred():
            status = ResultStatus.Warning
        elif self.create_file_signature_process.failed_output_error_code_occurred():
            status = ResultStatus.Error
        else:
            status = ResultStatus.Normal

        result = self.create_file_signature_process.output()

        self.show_signing_result(result, status)

    def on_pdf_conversion_process_finished(self) -> None:
        self.stop_waiting()
        if self.pdf_conversion_process.exitStatus() != QProcess.NormalExit or \
                self.pdf_conversion_process.exitCode() != 0:
            self.show_error(self.pdf_conversion_process.errorString(),
                            translate('text', 'Failed to open file'))
            return

        Utils.open_file(self.pdf_conversion_process.output_file_path(file_path=self.document,
                                                                     outdir=self.cache.cache_path()))

    def on_verify_file_process_finished(self) -> None:
        self.stop_waiting()
        if self.verify_file_process.exitStatus() != QProcess.NormalExit:
            self.show_error(self.verify_file_process.errorString(),
                            translate('text', 'Failed to verify file'))
            return

        if self.verify_file_process.timestamp_error_occurred():
            status = ResultStatus.Warning
        elif self.verify_file_process.failed_output_error_code_occurred():
            status = ResultStatus.Error
        else:
            status = ResultStatus.Normal

        result = self.verify_file_process.output()

        self.show_verification_result(result, status)

    def setup_certificates(self, certificates: List[Certificate]) -> None:
        self.comboBoxCertificates.clear()
        fprint = self.user_settings.fingerprint
        for certificate in sorted(certificates, key=lambda cert: cert.subject):
            self.comboBoxCertificates.addItem(certificate.subject, certificate.fingerprint)
            if fprint and certificate.fingerprint == fprint:
                self.comboBoxCertificates.setCurrentText(certificate.subject)

    def parse_documents_and_signatures(self, files: List[str]) -> Tuple[List[str], List[str]]:
        documents = []
        signatures = []
        for file in files:
            mimetype = Utils.file_mimetypes(file)
            if mimetype in self.sigmimetypes:
                signatures.append(file)
            else:
                documents.append(file)
        return documents, signatures

    def validate_file(self) -> None:
        if not self.document:
            self.show_critical_error(translate('text', 'No document selected to verify or create a digital signature.'))
        if Utils.file_is_exists(self.document):
            self.labelDocumentPath.setText(self.document)
            self.labelDocumentPathSign.setText(self.document)
            self.update_signature_name()
        else:
            self.show_critical_error(translate('text', 'The document does not exist'))

    def validate_signature(self) -> None:
        if self.signature and Utils.file_is_exists(self.signature):
            self.labelSignaturePath.setText(self.signature)
        else:
            self.labelSignaturePath.setText(translate('text', 'The built-in electronic signature will be verified.'))

    def maybe_verify(self) -> None:
        if self.signature:
            self.verify()

    def enable_ui(self) -> None:
        self.tabWidget.setEnabled(True)
        self.pushButtonSign.setEnabled(self.comboBoxCertificates.count() > 0)

    def disable_ui(self) -> None:
        self.tabWidget.setDisabled(True)

    def enable_controls(self) -> None:
        self.pushButtonDocumentView.setEnabled(True)
        self.pushButtonDocumentViewSign.setEnabled(True)
        self.pushButtonGenerateName.setEnabled(True)
        self.pushButtonSelectSignature.setEnabled(True)
        self.pushButtonVerifySignature.setEnabled(True)
        self.comboBoxCertificates.setEnabled(True)
        self.checkBoxIncludeTimestamp.setEnabled(True)
        self.lineEditSignatureFileName.setEnabled(True)
        self.pushButtonSign.setEnabled(self.comboBoxCertificates.count() > 0)

    def disable_controls(self) -> None:
        self.pushButtonSign.setDisabled(True)
        self.pushButtonDocumentView.setDisabled(True)
        self.pushButtonDocumentViewSign.setDisabled(True)
        self.pushButtonGenerateName.setDisabled(True)
        self.pushButtonSelectSignature.setDisabled(True)
        self.pushButtonVerifySignature.setDisabled(True)
        self.comboBoxCertificates.setDisabled(True)
        self.checkBoxIncludeTimestamp.setDisabled(True)
        self.lineEditSignatureFileName.setDisabled(True)

    def view(self) -> None:
        try:
            if Utils.file_mimetypes(self.document) in self.global_settings.pdf_conversion_mimetypes:
                if not Utils.create_directory(self.cache.cache_path()):
                    raise Exception(self.tr('Failed to create directory "{path}"').format(path=self.__cache_path))

                self.pdf_conversion_process.setup_arguments(file_path=self.document, outdir=self.cache.cache_path())
                self.start_waiting(InteractiveProcessDialog(process=self.pdf_conversion_process, parent=self))
                self.pdf_conversion_process.start()
            else:
                Utils.open_file(self.document)
        except Exception as exception:
            self.show_error(translate('text', 'Failed to open document'),
                            translate('text', 'Failed to open document'),
                            details=str(exception))

    def verify(self) -> None:
        self.start_waiting(CprocspProcessDialog(process=self.verify_file_process, parent=self))
        self.clear_verification_result()
        self.verify_file_process.setup_arguments(file_path=self.document, sig_name=self.signature)
        self.verify_file_process.start()

    def sign(self) -> None:
        self.start_waiting(CprocspProcessDialog(process=self.create_file_signature_process, parent=self))
        self.clear_sign_result()
        dirname = os.path.dirname(self.document)
        filename = self.lineEditSignatureFileName.text()
        filepath = "{dir}/{file}".format(dir=dirname, file=filename)
        fprint = self.comboBoxCertificates.currentData()
        self.save_fingerprint(fprint)
        self.create_file_signature_process.setup_arguments(file_path=self.document,
                                                           sig_name=filepath,
                                                           fingerprint=fprint,
                                                           time_address=self.global_settings.time_address
                                                           if self.include_timestamp() else None)

        self.create_file_signature_process.start()

    def save_fingerprint(self, fprint: str) -> None:
        self.user_settings.fingerprint = fprint

    def update_signature_name(self) -> None:
        name = self.generate_signature_name(self.document)
        self.lineEditSignatureFileName.setText(name)

    def generate_signature_name(self, path: str) -> str:
        date = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
        base = os.path.basename(path)
        return "{base}_{date}.sig".format(base=base, date=date)

    def include_timestamp(self) -> bool:
        return self.checkBoxIncludeTimestamp.isChecked()

    def select(self) -> None:
        dirname = os.path.dirname(self.document)
        dialog = QFileDialog(self, translate('text', 'Select signature'), dirname)
        dialog.setNameFilters([translate('text', 'Signature files (*.p7s *.asc *.sig *.pgp *.gpg)'),
                               translate('text', 'All files (*)')])
        if dialog.exec():
            self.signature = dialog.selectedFiles()[0]
            self.labelSignaturePath.setText(self.signature)
            self.clear_verification_result()

    def show_verification_result(self, result: str, status: ResultStatus) -> None:
        self.show_result(self.textBrowserResult, result, status)

    def show_signing_result(self, result: str, status: ResultStatus) -> None:
        self.show_result(self.textBrowserResultSign, result, status)

    def show_result(self, widget: QTextBrowser, result: str, status: ResultStatus) -> None:
        widget.setPalette(self.status_palette(status))
        widget.setText(result)

    def clear_verification_result(self) -> None:
        self.textBrowserResult.clear()
        self.textBrowserResult.setPalette(QGuiApplication.palette())

    def clear_sign_result(self) -> None:
        self.textBrowserResultSign.clear()
        self.textBrowserResultSign.setPalette(QGuiApplication.palette())

    def status_palette(self, status: ResultStatus) -> QPalette:
        if status == ResultStatus.Normal:
            return self.success_palette
        elif status == ResultStatus.Warning:
            return self.warning_palette
        else:
            return self.error_palette

    def start_waiting(self, dialog: InteractiveProcessDialog) -> None:
        self.waiting_dialog = dialog
        self.waiting_dialog.setWindowTitle(translate('text', 'Processing'))
        self.disable_controls()
        self.waiting_timer.start()

    def stop_waiting(self) -> None:
        self.waiting_timer.stop()
        self.close_waiting_dialog()
        self.enable_controls()

    def show_waiting_dialog(self) -> None:
        if not self.is_cryptopro_ui_running() and self.waiting_dialog:
            self.waiting_dialog.exec()

    def close_waiting_dialog(self) -> None:
        if self.waiting_dialog:
            self.waiting_dialog.close()
        self.waiting_dialog = None

    def show_help(self) -> None:
        url = 'qthelp://flyhelp.ru/doc/fly-csp/fly-csp.html'
        path = '/usr/share/doc/fly/qthelp/ru/flyhelp.qhc'
        QProcess.startDetached("assistant -collectionFile {path} -showUrl {url}".format(path=path, url=url))

    def show_error(self, error: str, title: str, details: str = None) -> None:
        self.stop_waiting()
        msgbox = QMessageBox(self)
        msgbox.setWindowTitle(title)
        msgbox.setText(error)
        if details:
            msgbox.setDetailedText(details)
        msgbox.exec()

    def show_critical_error(self, error: str, title: str = None, details: str = None) -> None:
        self.show_error(error, title or translate('text', 'Critical error'), details)
        self.exit()

    def clear_cache(self) -> None:
        self.cache.clear_cache()

    def exit(self) -> None:
        self.clear_cache()
        sys.exit(0)

    def is_cryptopro_ui_running(self):
        uid = os.getuid()
        for proc in psutil.process_iter():
            try:
                if proc.name() == "xcpui_app" and proc.uids().real == uid:
                    return True
            except psutil.Error:
                pass
        return False

    def validate_system(self) -> None:
        try:
            current_max_level = Utils.get_mac_level()
        except:
            raise Exception(self.tr('Error getting mandated level'))

        available_max_level = self.global_settings.max_mac_level
        if current_max_level > available_max_level:
            message = self.tr('The use of csp service at mandated '
                              'level {current}(>{available}) is prohibited').format(current=current_max_level,
                                                                                    available=available_max_level)
            raise Exception(message)

        if not Utils.file_is_exists('/opt/cprocsp/bin/amd64/cryptcp'):
            raise Exception(self.tr('Crypto provider not installed'))

        if not Utils.file_is_exists('/opt/cprocsp/bin/amd64/certmgr'):
            raise Exception(self.tr('Certificate manager not installed'))
