#!/usr/bin/python3
# -*- coding: utf-8 -*-

from functools import wraps
from typing import Optional, List

from PyQt5.QtCore import QSettings, QCoreApplication

from astra_csp.utils import Utils

translate = QCoreApplication.translate


def check_status(function):
    @wraps(function)
    def wrapper(self, *args, **kwargs):
        return_value = function(self, *args, **kwargs)

        if self.status == QSettings.AccessError:
            message = translate('Settings',
                                'An access error occurred to "{file}" file').format(file=self.file_path)
            raise Exception(message)
        elif self.status == QSettings.FormatError:
            message = translate('Settings',
                                'File format error "{file}"').format(file=self.file_path)
            raise Exception(message)
        elif self.status != QSettings.NoError:
            message = translate('Settings',
                                'Unknown error reading configuration "{file}" file').format(file=self.file_path)
            raise Exception(message)

        return return_value

    return wrapper


class Settings:
    def __init__(self, settings: QSettings):
        self._settings = settings

    @property
    def file_path(self) -> str:
        return self._settings.fileName()

    @property
    def status(self):
        return self._settings.status()


class GlobalSettings(Settings):
    __max_mac_level_key = 'MaxMacLevel'
    __time_address_key = 'TimeAddress'
    __max_csp_time_waiting_key = 'MaxCspTimeWaiting'
    __pdf_conversion_mime_types_key = 'PdfConversionMimeTypes'

    def __init__(self) -> None:
        super().__init__(QSettings('/etc/astra-csp/fly-csp.conf', QSettings.NativeFormat))
        self.__validate()

    def __validate(self) -> None:
        if not Utils.file_is_exists(self.file_path):
            message = translate('Settings',
                                'Configuration file "{file}" not exists').format(file=self.file_path)
            raise Exception(message)

        self.__check_parameter(self.__max_mac_level_key, self.max_mac_level)
        self.__check_parameter(self.__time_address_key, self.time_address)

    @property
    @check_status
    def max_mac_level(self) -> Optional[int]:
        value = self._settings.value(self.__max_mac_level_key)
        return int(value) if value else None

    @property
    @check_status
    def time_address(self) -> Optional[str]:
        return self._settings.value(self.__time_address_key)

    @property
    @check_status
    def max_csp_time_waiting(self) -> Optional[int]:
        value = self._settings.value(self.__max_csp_time_waiting_key)
        return int(value) if value else None

    @property
    @check_status
    def pdf_conversion_mimetypes(self) -> List[str]:
        return self._settings.value(self.__pdf_conversion_mime_types_key, [])

    def __check_parameter(self, parameter: str, value) -> None:
        if value is None:
            message = translate('Settings',
                                'Parameter "{parameter}" is not set in "{file}" configuration file') \
                .format(parameter=parameter, file=self.file_path)

            raise Exception(message)


class UserSettings(Settings):
    __fingerprint_key = 'LastFingerprint'

    def __init__(self) -> None:
        super().__init__(QSettings(QSettings.NativeFormat, QSettings.UserScope, 'rusbitech', 'fly-csp'))

    @property
    def fingerprint(self) -> Optional[str]:
        return self._settings.value(self.__fingerprint_key)

    @fingerprint.setter
    def fingerprint(self, value: str) -> None:
        self._settings.setValue(self.__fingerprint_key, value)
        self._settings.sync()
